import 'dart:async';

import 'package:lightweight_bloc/src/bloc_observer.dart';

abstract class Bloc<T> extends Stream<T> {
  static final _blocObserver = BlocObserver();

  final _stateController = StreamController<T>.broadcast();

  T _state;

  List<T> _previousStates;

  Bloc({T initialState}) {
    assert(initialState != null || this.initialState != null);

    _state = initialState ?? this.initialState;

    _previousStates = List<T>();
    _previousStates.add(_state);
  }

  T get initialState => null;

  T get state => _state;

  void init();

  void update(T newState) {
    if (!_stateController.isClosed) {
      if (_state != newState) {
        _blocObserver.invokeCallbacks(this, state, newState);
        _state = newState;
        _stateController.add(newState);
        _previousStates.add(_state);
      }
    }
  }

  @override
  StreamSubscription<T> listen(void onData(T event),
      {Function onError, void onDone(), bool cancelOnError}) {
    return _stateStream.listen(
      onData,
      onError: onError,
      onDone: onDone,
      cancelOnError: cancelOnError,
    );
  }

  Stream<T> get _stateStream async* {
    yield state;
    yield* _stateController.stream;
  }

  void dispose() {
    _stateController.close();
  }

  void undo() {
    _previousStates.removeLast();
    update(_previousStates.last);
  }
}
